from ..dependencies import import_or_install

try:
    import magic
except ImportError:  # pragma nocover
    magic = import_or_install("magic", "python-magic")  # pip install python-magic
# end try
import mimetypes


def guess_extension(mime):
    """Shortcut for getting extension to a given mime string.
    The parameter mime can be None"""
    return mimetypes.guess_extension(type=mime or "")


def get_file_mime(file_path=None, file_url=None):
    """
    Shortcut to get the mime from either

    :param file_path:
    :param file_url:
    :return:
    """
    if file_url:
        url = file_url
    elif file_path:
        from urllib.request import pathname2url
        url = pathname2url(file_path)
    else:
        raise AttributeError("Neither URL (file_url) nor local path (file_path) given.")
    mime = magic.from_file(file_path).decode("utf-8")
    return mime


def get_byte_mime(bytes):
    """
    Shortcut to get a mime from bytes in a variable.

    :param bytes:
    :return:
    """
    magic.from_buffer(bytes, mime=True).decode("utf-8")


def get_file_suffix(file_path=None, file_url=None):
    """
    This calls `get_file_mime()` to get the mime, and then calls `guess_extension()`.
    """
    mime = get_file_mime(file_path=file_path, file_url=file_url)
    return guess_extension(mime)



# -*- coding: utf-8 -*-
from luckydonaldUtils.logger import logging
import typing

__author__ = 'luckydonald'
logger = logging.getLogger(__name__)


class Launcher(object):
    """
    Allow ``flask``'s auto-reload to survive syntax errors and similar exceptions.
    Those will be displayed in the browser (if `debug` is on), or a generic message is provided.

    Also it will check if there is a debugger connected, and disable auto-reload.

    Finally it will detect if ``main.socketio`` exists, and launch that instead.
    (See https://flask-socketio.readthedocs.io/en/latest/#embedded-server for more info about that)

    ---

    Usage:

    Instead of

        >>> from somewhere.main import app
        >>>
        >>> if __name__ == "__main__":
        >>>     try:  # https://stackoverflow.com/a/338391/3423324
        >>>         import inspect
        >>>         debugger_connected = any(frame[1].endswith("pydevd.py") for frame in inspect.stack())
        >>>     except Exception:
        >>>         debugger_connected = False
        >>>     # end try
        >>>     auto_reload = not debugger_connected
        >>>     app.run(host='0.0.0.0', port=8080, debug=True, use_reloader=auto_reload)
        >>> # end if

    you simply do

        >>> from luckydonaldUtils.flasks.launcher import Launcher
        >>>
        >>> with Launcher(__name__, host='0.0.0.0', port=8080, debug=True) as l:
        >>>     from somewhere import main
        >>>     l(main)
        >>> # end with
        >>>
        >>> app = l.app

    """

    def __init__(self, name, *args, **kwargs):
        """Run the SocketIO/Flask web server.
        :param name: The ``__name__`` variable.
        :param host: The hostname or IP address for the server to listen on.
                     Defaults to 127.0.0.1.
        :param port: The port number for the server to listen on. Defaults to
                     5000.
        :param debug: ``True`` to start the server in debug mode, ``False`` to
                      start in normal mode.
        :param use_reloader: ``True`` to forcefully enable the Flask reloader, ``False``
                             to forcefully disable it, ``None`` to disable automatically
                             if a debugger is connected. Defaults to ``None``.
        :param templates_use_reloader: ``None`` to enable the jijnja2 reloader only
                                       if no debugger is connected. ``True`` to
                                       forcefully enable it, and ``False`` to
                                       forcefully disable it. Defaults to ``None``.
        :param extra_files: A list of additional files that the Flask
                            reloader should watch. Defaults to ``None``.
        :param log_output: If ``True``, the server logs all incomming
                           connections. If ``False`` logging is disabled.
                           Defaults to ``True`` in debug mode, ``False``
                           in normal mode. Unused when the threading async
                           mode is used.
        :param kwargs: Additional web server options. The web server options
                       are specific to the server used in each of the supported
                       async modes. Note that options provided here will
                       not be seen when using an external web server such
                       as gunicorn, since this method is not called in that
                       case.
        """
        if name == '__main__':
            logging.add_colored_handler(level=logging.DEBUG)
        # end if
        self.name = name
        self.args = args
        self.kwargs = kwargs
        self.main = None
        self.did_fail = False
        self.exception = None

    # end def

    def __enter__(self):
        return self

    # end def

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            logger.exception('could not load app code:')
            self.main = None
            self.exception = (exc_type, exc_value, traceback)
        # end if
        return True  # suppress the exception  # https://docs.python.org/3/whatsnew/2.6.html#writing-context-managers

    # end def

    def __call__(self, main):
        self.main = main

    # end def

    @property
    def app(self):
        """
        :rtype:
        """
        if self.name == "__main__":
            # "__main__" means, this python file is called directly.
            # not to be confused with "main" (because main.py) when called from from nginx

            if self.main is None:
                # loading failed due to Syntax Errors or similar.
                self.did_fail = True
                self.fake_main_app()
            # end if

            debugger_connected = self.check_debugger()

            from flask.helpers import get_debug_flag
            do_debug = self.kwargs.pop('debug', get_debug_flag() or debugger_connected)
            self.kwargs.setdefault('use_debugger', debugger_connected)
            self.kwargs.setdefault('use_reloader', do_debug)
            self.kwargs.setdefault('templates_use_reloader', do_debug)

            N, T, F = None, True, False
            auto_reload, auto_reload_templates = {
                N: {
                    #  (flask, templates)
                    N: (not debugger_connected, not debugger_connected),
                    T: (not debugger_connected, T),  # flask, templates
                    F: (not debugger_connected, F),  # flask, templates
                },
                T: {
                    #  (flask, templates)
                    N: (T, T),  # flask, templates
                    T: (T, T),  # flask, templates
                    F: (T, F),  # flask, templates
                },
                F: {
                    #  (flask, templates)
                    N: (F, F),  # flask, templates
                    T: (F, F),  # flask, templates
                    F: (F, F),  # flask, templates
                },
            }[self.kwargs.pop('use_reloader')][self.kwargs.pop('templates_use_reloader')]

            logger.info('Debugger detected: {c}; Auto reload flask/templates {f}/{t}; Debug: {d}'.format(
                c=debugger_connected, f=auto_reload, t=auto_reload_templates, d=do_debug,
            ))

            # enable template auto reload
            if auto_reload_templates:
                self.main.app.jinja_env.auto_reload = True
                self.main.app.config['TEMPLATES_AUTO_RELOAD'] = True
            # end def

            self.kwargs.setdefault('host', '0.0.0.0')
            self.kwargs.setdefault('port', 8080)
            self.kwargs.setdefault('debug', do_debug)
            self.kwargs.setdefault('use_reloader', auto_reload)
            self.kwargs.setdefault('extra_files', [])

            if self.did_fail:
                from pathlib import Path
                import sys, os
                main_file = os.path.abspath(sys.modules['__main__'].__file__)
                print(main_file)
                self.kwargs['extra_files'].extend([x.absolute() for x in Path(main_file).rglob("*.py")])
                self.kwargs['debug'] = True
            # end if
            self.run_app()
        return self.main

    # end def

    def run_app(self):
        logger.debug("launching app with arguments: {!r}, {!r}".format(self.args, self.kwargs))
        if hasattr(self.main, 'socketio'):
            self.main.socketio.run(self.main.app, *self.args, **self.kwargs)
        else:
            self.main.app.run(*self.args, **self.kwargs)
        # end if

    # end def

    def fake_main_app(self, text='App not started', status=500):
        # loading failed due to Syntax Errors or similar.
        from flask import Flask
        self.main = lambda: 0  # fake main
        self.main.app = Flask(__name__)  # simple app
        self.main.app.debug = self.kwargs.get('debug', False)
        self.main.app.errorhandler(500)(self.error_handler(text, status))
        self.main.app.errorhandler(404)(self.error_handler(text, status))
    # end def

    def error_handler(self, text='App not started', status=500):
        from werkzeug.debug.tbtools import Traceback
        from werkzeug.debug import DebuggedApplication
        from flask import request

        def error_handler_inner_simple(e):
            return text, status

        # end def

        if not self.main.app.debug:
            return error_handler_inner_simple
        # end if

        d = DebuggedApplication(None)

        def error_handler_inner(e):
            if request.args.get('__debugger__') == 'yes' and request.args.get('cmd') == "resource":
                return d.get_resource(request, request.args.get('f'))
            # end if

            if self.exception:
                exc_type, exc_value, traceback = self.exception
                tb = Traceback(exc_type, exc_value, traceback)
                html = tb.render_full(evalex=False, secret=None, evalex_trusted=False)
                html = html.replace('</head>', '<style>.footer,.pastemessage{display:none}</style></head>')
                return html, status
            # end if
            return error_handler_inner_simple(e)

        # end def
        return error_handler_inner
    # end def

    @staticmethod
    def check_debugger():
        # check for debugger, as auto reloading with syntax errors always triggers the debugger,
        # which will jump to some random point where the code breaks, which is annoying.
        # If a debugger is connected, don't reload.

        # noinspection PyBroadException
        try:
            # https://stackoverflow.com/a/338391/3423324
            import inspect
            return any(frame[1].endswith("pydevd.py") for frame in inspect.stack())
        except Exception:
            return False
        # end try
    # end def
# end class



# -*- coding: utf-8 -*-
__author__ = 'luckydonald'

from luckydonaldUtils.logger import logging  # pip install luckydonald-utils
logger = logging.getLogger(__name__)


def iter_with_i(iterator, i_start = 0):
    """
    Yields a tuple of the iterator result and an i(iterator_result, i),
    i incrementing each time.

    Example:
        >>> some_list = ["a","b"]

        >>> for iterator_result, i in iter_with_i(some_list)
        ... 	print("iterator is {}, i is {}".format(iterator_result, i))
        iterator is a, i is 0
        iterator is b, i is 1

    :param iterator: Your iterable object (lists etc.)
    :keyword i_start: Where to start counting. Default is `0`
    """
    i = i_start
    for iterator_result in iterator:
        yield (iterator_result, i)
        i += 1




# -*- coding: utf-8 -*-
from luckydonaldUtils.files import logger
from luckydonaldUtils.logger import logging

__author__ = 'luckydonald'
logger = logging.getLogger(__name__)


def do_a_filename(input_file_name):
    """
    This is a bad try to make file names better, replacing various characters with best fitting ascii ones.
    This does only include a basic set of replacements.
    Note: In no way the output should be considered safe to use! Not all non-ascii characters are replaced!
    """
    output_file_name = input_file_name
    for replacer in [(":", ""), ("?", ""), ("*", ""), ("”", "\""), (">", ""), ("<", ""), ("|", "-"), ("\\", " "),
                     ("/", " ")]:
        output_file_name = output_file_name.replace(replacer[0], replacer[1])
    logger.debug("Filename '{old_filename}' is now '{new_filename}'.".format(old_filename=input_file_name,
                                                                             new_filename=output_file_name))
    return output_file_name



# -*- coding: utf-8 -*-
try:
    from django.conf import settings
    from django.core.exceptions import ImproperlyConfigured, MiddlewareNotUsed
    from django.http import HttpResponseForbidden
except ImportError:  # pragma nocover
    from ...dependencies import import_or_install

    settings = import_or_install("django.conf.settings", "django")
    ImproperlyConfigured = import_or_install("django.ore.exceptions.ImproperlyConfigured", "django")
    MiddlewareNotUsed = import_or_install("django.ore.exceptions.MiddlewareNotUsed", "django")
    HttpResponseForbidden = import_or_install("django.http.HttpResponseForbidden", "django")
# end try
from ...logger import logging  # pip install luckydonald-utils
from ...dependencies import import_or_install

__author__ = 'luckydonald'
logger = logging.getLogger(__name__)

IP = import_or_install("IPy.IP", "IPy")  # pip install IPy


class AllowFromIPMiddleware(object):
    """
    Allow only given IPs/CIDRs to access, else raise a `Http404` error. Is ignored when `DEBUG` is `True`.
    Checks ranges with https://pypi.python.org/pypi/IPy/.

    Set `ALLOW_FROM` to a list of IPs/CIDRs strings. To disable the Middleware, set it to `None`.

    Example:
    The *`settings.py`*
    ```
    ALLOW_FROM = ["8.8.8.8", "134.169.0.0/16"]
    ```

    To mimic the Apache syntax:
    > Allow from 134.169.
    One would write
    >>> ALLOW_FROM = ["134.169"]
    or
    >>> ALLOW_FROM = ["134.169.0.0/16"]
    in the `settings.py`.
    """
    buffered_ALLOW_FROM = []

    def __init__(self):
        try:
            allow_from = settings.ALLOW_FROM
        except AttributeError:
            raise ImproperlyConfigured('ALLOW_FROM setting does not exist.')
        if allow_from is None:
            raise MiddlewareNotUsed("ALLOW_FROM is None.")
        if not isinstance(allow_from, (list,tuple)):
            raise ImproperlyConfigured("ALLOW_FROM setting isn't a list of IP/CIDR strings.")
        if len(allow_from) == 0:
            raise ImproperlyConfigured("ALLOW_FROM is an empty list.")
        for ip in allow_from:
            self.buffered_ALLOW_FROM.append(IP(ip))
        #end for
    #end def __init__

    def process_request(self, request):
        remote_addr = request.META.get('HTTP_X_REAL_IP', request.META.get('REMOTE_ADDR', None))
        if any([remote_addr in ip for ip in self.buffered_ALLOW_FROM]):
            return  # allow
        if settings.DEBUG:
            logger.warn("The IP {remote_addr} should not be allowed, but is served anyway because DEBUG=True.".format(remote_addr=remote_addr))
            return  # allow
        return HttpResponseForbidden("Your ip {ip} is not allowed.".format(ip=remote_addr))   # forbid
            #end if DEBUG
        #end if
    #end def
#end class


# -*- coding: utf-8 -*-
from datetime import datetime
from . import py2

__author__ = 'luckydonald'


def to_jsonable_dict(skip_keys=list()):
    dict(to_jsonable_dict(skip_keys=skip_keys))

if py2:
    def to_jsonable_dict_iterator(dict, skip_keys=list()):
        for key, value in dict.iteritems():
            if key in skip_keys:
                continue
            if isinstance(value, datetime):
                yield (key, str(value.isoformat()))
            else:
                yield (key, value)
else:
    def to_jsonable_dict_iterator(dict, skip_keys=list()):
        for key, value in dict.items():
            if key in skip_keys:
                continue
            if isinstance(value, datetime):
                yield (key, str(value.isoformat()))
            else:
                yield (key, value)


# -*- coding: utf-8 -*-
import sys
import os
from luckydonaldUtils.logger import logging

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)


def mkdir_p(path):
    """
    like mkdir -p
    Creates a folder with all the missing parent folders.
    """
    import errno
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise
            # end try


# end def

"""
def open_folder(folder_path)
"""  # for different Platforms
if sys.platform == 'darwin':  # Mac OS
    logger.debug('platform "darwin", Mac OS X. Using `open` to open files/folders.')


    def open_folder(folder_path):
        """tries to open a folder in your system's browser"""
        import subprocess
        subprocess.check_call(['open', '--', folder_path])


    # end def

    def open_file_folder(file_path):
        """tries to open a folder, and select the given file in your system's browser"""
        import subprocess
        subprocess.check_call(['open', '-R', '--', file_path])
        # end def

elif sys.platform == 'linux2':  # linux, hopefully has gnome?
    logger.debug('platform "linux2", hopefully has gnome. Using `gnome-open` to open files/folders.')


    def open_folder(folder_path):
        """tries to open a folder in your system's browser"""
        import subprocess
        subprocess.check_call(['gnome-open', '--', folder_path])  # untested.


    # end def

    def open_file_folder(file_path):
        """tries to open a folder, and select the given file in your system's browser"""
        import subprocess
        open_folder(os.path.dirname(file_path))
        # end def
elif sys.platform == 'win32':  # windows
    logger.debug("platform win32, windows. Using `explorer` to open files/folders.")


    def open_folder(folder_path):
        """tries to open a folder in your system's browser"""
        import subprocess
        subprocess.check_call(['explorer', folder_path])


    # end def

    def open_file_folder(file_path):
        """tries to open a folder, and select the given file in your system's browser"""
        import subprocess
        try:
            subprocess.check_call(['explorer', '/select,"%s"' % file_path])
        except:
            open_folder(os.path.dirname(file_path))
            # end def
# end if



# -*- coding: utf-8 -*-
import re

__author__ = 'luckydonald'


REPO_NAME_REGEX = re.compile("([a-zA-Z0-9_.-]+)")

AT_USERNAME_REGEX = re.compile(  # https://regex101.com/r/eJ6cA4/
    "(?:\s|^|\.|,|:|\?|!|¡|¿)@(?!-)(?P<user>(?:[a-zA-Z0-9]|-(?!-))+[a-zA-Z0-9])(?![^\x00-\x7F])(?:$|[^a-zA-Z0-9\-_])"
)

SIMPLE_URL_REGEX = re.compile(
    r"(?:https?://(?:www\.)?|www\.|^|(?<=\s))github.com/[^\s]+")  # https://regex101.com/r/gH8xM9/4
""" :const:`SIMPLE_URL_REGEX`: A very simple url regex """

FILE_URL_REGEX = re.compile(  # https://regex101.com/r/qK6uN9/15
    "\\b(?P<url>(?P<protocol>https?://)?(?P<www>www\.)?github.com/(?P<user>[a-zA-Z0-9_-]+)/(?P<repo>[a-zA-Z0-9_.-]+)/?"
    "(?P<path>(?:(?P<kind>blob|tree)/(?P<branch>[a-zA-Z0-9_%-]+))/?(?P<file>[^\s#]*)/?)?)"
    "(?:#(?P<hash>[-;/?:@&=+$,_.!~*'()%0-9a-zA-Z]*))?(?:(?=\s)|$)", flags=re.IGNORECASE
)
"""
:const:`FILE_URL_REGEX`: Matching github urls with the following groups:
  - url: the complete url
      - protocol: "https://" or "http://" or empty/non-existent
      - user: git user or organisation
      - repo: the repository
      - path: When existent, this is not the project page (root of master)
          - kind: blob or tree
          - branch: the name of the branch (kind=tree), or the commit hash (kind=blob)
        - file: the rest of the filepath (from root of that branch, can be empty)
      - hash: Can be non-existent or empty. Everything behind the '#'
"""



# -*- coding: utf-8 -*-
import logging as _logging
import sys

__author__ = 'luckydonald'
__all__ = ["logging", "ColoredFormatter", "ColoredStreamHandler", "LevelByNameFilter"]


class ColoredFormatter(_logging.Formatter):
    class Color(object):
        """
         utility to return ansi colored text.
         just to store the colors next to the function.
        """
        # Color codes: http://misc.flogisoft.com/bash/tip_colors_and_formatting

        def __init__(self, formatter):
            self.formatter = formatter

        # end def

        colors = {
            'default': 39,
            'black': 30,
            'red': 31,
            'green': 32,
            'yellow': 33,
            'blue': 34,
            'magenta': 35,
            'cyan': 36,
            'white': 37,
            'grey': 90,
            'bgred': 41,
            'bggrey': 100
        }
        mapping = {
            'INFO': 'default',
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'magenta',
            'DEBUG': 'grey',
            'SUCCESS': 'green'
        }

        color_prefix = '\033['

        def prepare_color(self, color_number):
            return ('%s%dm') % (self.color_prefix, color_number)

        # end def

        def colored(self, record):
            """
            adsasd
            """

            color = self.mapping.get(record.levelname, 'default')
            clr = self.colors[color]
            formatter = dict(
                all_off=self.prepare_color(0),  # Reset all attributes
                color_on=self.prepare_color(clr),  # Color as given/from lookup
                color_off=self.prepare_color(39),  # Default foreground color
                inverse_on=self.prepare_color(7),  # Reverse (invert the foreground and background colors)
                inverse_off=self.prepare_color(27),  # Reset reverse
                background_off=self.prepare_color(49),  # Default background color
                file_color_on=self.prepare_color(94),  # Light blue
            )
            lines = []

            # log level
            level = "{level:8}".format(level=record.levelname)
            level_filler = "{:{}}".format("", len(level))

            # file/function name
            filepart = record.name if record.name else ""
            filepart += "." + record.funcName if record.funcName != "<module>" else ""

            # date
            timestamp = " " + record.asctime if record.asctime else ""
            timestamp_filler = " " * len(timestamp)

            # Process / Thread names
            process_thread_part = process_thread_part_filler = ""
            has_process = hasattr(record, "processName") and record.processName != "MainProcess"
            has_thread = hasattr(record, "threadName") and record.threadName != "MainThread"

            if has_process:
                process_thread_part += "{inverse_on}{file_color_on}{thread}{inverse_off}".format(
                    thread=record.processName, **formatter)
            # end if
            if has_process and has_thread:
                process_thread_part += " ".format(**formatter)
            # end if
            if has_thread:
                process_thread_part += "{inverse_on}{file_color_on}{process}{inverse_off}".format(
                    process=record.threadName, **formatter)
            # end if
            if has_process or has_thread and len(timestamp) > 1:
                # inject the formatting here, as empty formatting without text would break
                process_thread_part_filler = " " * len(process_thread_part)
                process_thread_part = "{file_color_on}{inverse_on}{process_thread_part}{inverse_off}".format(
                    process_thread_part=process_thread_part, **formatter
                )
                # abuse {date} to contain a space for us. Because a blue colored space is still a space.
                timestamp += " "  # so the file don't immediatly follows after the date.
                timestamp_filler += " "
            # end if


            # original message
            lines_ = record.message.splitlines()
            first_line = True if len(lines_) > 1 else  None
            for line in lines_:
                if first_line is None:  # single line
                    lines.append(
                        "{color_on}{inverse_on}{level}{inverse_off}{color_on}{date}{color_off}{file_color_on}{process_thread_part} {file_color_on}{filepart}:{color_off} {color_on}{message}{color_off}{background_off}{all_off}".format(
                            filepart=filepart, level=level, message=line, date=timestamp,
                            process_thread_part=process_thread_part, **formatter))
                    break
                elif first_line:  # first line
                    lines.append(
                        "{color_on}{inverse_on}{level}{inverse_off}{color_on}{date}{color_off}{file_color_on}{process_thread_part} {file_color_on}{filepart}:{color_off} {all_off}".format(
                            filepart=filepart, level=level, message=line, date=timestamp,
                            process_thread_part=process_thread_part, **formatter))
                lines.append(
                    "{color_on}{inverse_on}{level_filler}{inverse_off}{color_off} {color_on}{message}{color_off}{background_off}{all_off}".format(
                        level_filler=level_filler, message=line, date=timestamp, date_filler=timestamp_filler,
                        process_thread_part=process_thread_part, process_thread_part_filler=process_thread_part_filler,
                        **formatter))
                first_line = False
            # end for
            return "\n".join(lines)
            # end def

    def __init__(self, date_formatter=None):
        super(ColoredFormatter, self).__init__(datefmt=date_formatter)
        self.color_instance = self.Color(self)

    def colored(self, record):
        return self.color_instance.colored(record)

    # end def


    def format(self, record):
        super(ColoredFormatter, self).format(record)
        # if record.threadName == "MainThread":
        # 	pass
        # part1 = self.firstpart.format(record)
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)
        else:
            record.asctime = ""
        s = self._fmt % record.__dict__  # py3: s = self.formatMessage(record)
        if record.exc_text:
            if s[-1:] != "\n":
                s += "\n"
            try:
                s = s + record.exc_text
            except UnicodeError:  # PYTHON 2, LOL!
                # Sometimes filenames have non-ASCII chars, which can lead
                # to errors when s is Unicode and record.exc_text is str
                # See issue 8924.
                # We also use replace for when there are multiple
                # encodings, e.g. UTF-8 for the filesystem and latin-1
                # for a script. See issue 13232.
                s = s + record.exc_text.decode(sys.getfilesystemencoding(), 'replace')
        if hasattr(record, "stack_info") and record.stack_info:  # py2 doesn't have .stack_info
            if s[-1:] != "\n":
                s += "\n"
            s = s + record.stack_info  # py3: self.formatStack()
        record.message = s
        return self.colored(record)

    # end def

    def usesTime(self):
        return bool(self.datefmt)
        # end def


# end class


class ColoredStreamHandler(_logging.StreamHandler):
    """
    Like the normal StreamHandler,
    but it automatically sets
    `self.formatter = ColoredFormatter()`
    """

    def __init__(self, stream=None, date_formatter=None):
        super(ColoredStreamHandler, self).__init__(stream)
        self.formatter = ColoredFormatter(date_formatter=date_formatter)


# noinspection PyProtectedMember,PyProtectedMember
class _LoggingWrapper(object):
    SUCCESS = 25  # between WARNING and INFO

    def __init__(self):
        _logging.addLevelName(self.SUCCESS, 'SUCCESS')

    def getLoglevelInt(self, level_string):
        """
        You provide a String, and get a level int
        :param level_string: The level.
        :type  level_string: str
        :return: level
        :rtype : int
        :raises KeyError: if the level does not exists.
        """
        if isinstance(level_string, int):
            return level_string
        # end if
        try:
            return {
                # as names:
                "NOTSET": _logging.NOTSET,
                "DEBUG": _logging.DEBUG,
                "INFO": _logging.INFO,
                "SUCCESS": self.SUCCESS,
                "WARNING": _logging.WARNING,
                "WARN": _logging.WARN,  # = WARNING
                "ERROR": _logging.ERROR,
                "FATAL": _logging.FATAL,  # = CRITICAL
                "CRITICAL": _logging.CRITICAL,
            }[level_string]
        except KeyError:
            try:
                return int(level_string)
            except ValueError:
                pass
            # end try
            raise  # key not known, and is no integer either.
            # end try

    # end def

    def __call__(self, logger_name):
        """
        alias to logger.getLogger(logger_name)
        :param logger_name:
        :return: self.getLogger(logger_name)
        """
        return self.getLogger(logger_name)

    # end def

    def add_colored_handler(self, logger_name=None, stream=None, level=None, date_formatter=None, filter=None):
        """
        Register a logger handler to colorfull print the messages.

        If stream is specified, the instance will use it for logging output; otherwise, sys.stdout will be used.

        If you supply a date_formatter, there will also be printed a date/time for the logged messages.
        Uses python `time.strftime` time formating, see https://docs.python.org/library/time.html#time.strftime

        :keyword logger_name: the name of the logger you want to register the printing to.
                              Probably you should use __name__ , to use your package's logger,
                              "root" will force all loggers to output.
        :type    logger_name: str

        :keyword stream: An output stream. Default: sys.stdout

        :keyword date_formatter: Apply a format for time output. If `None` is given, no time gets printed.
                               Something like "%Y-%m-%d %H:%M:%S". Uses python `time.strftime` time formating,
                               see https://docs.python.org/library/time.html#time.strftime
        :type    date_formatter: str

        :keyword filter: A filter applied to the handler.

        :return: None
        """
        logger = self.getLogger(logger_name)  # __name__
        if stream is None:
            import sys
            stream = sys.stdout
        # end if
        handler = ColoredStreamHandler(stream=stream, date_formatter=date_formatter)
        if filter:
            handler.addFilter(filter)
        # end if
        logger.addHandler(handler)
        if level:
            logger.setLevel(level)
        # end if
        return logger

    # end def

    def test_logger_levels(self, name=__name__, force_all_levels=True):
        logger = self.getLogger(name)
        logger_level = logger.getEffectiveLevel()
        if force_all_levels:
            logger.setLevel(logging.DEBUG)
        logger.debug('level debug')
        logger.info('level info')
        logger.success('level success')
        logger.warning('level warning')
        logger.error('level error')
        logger.critical('level critical')
        if force_all_levels:
            logger.setLevel(logger_level)
            # end if

    # end def

    def getLogger(self, name=None):
        """
        Adds the .success() function to the logger, else it is same as logger.getLogger()
        :param logger: a logging.getLogger() logger.
        :return:
        """
        logger = _logging.getLogger(name)
        logger.SUCCESS = self.SUCCESS
        setattr(logger, "success", lambda message, *args: logger._log(self.SUCCESS, message, args))
        return logger

    if sys.version < "3":
        def success(self, msg, *args, **kwargs):
            """
            Log 'msg % args' with severity 'SUCCESS'.

            To pass exception information, use the keyword argument exc_info with
            a true value.

            logger.debug("Houston, we landed in the %s", "moon", exc_info=False)
            """
            self._success(msg, *args, **kwargs)
    else:
        from .py3 import success

    def _success(self, msg, *args, **kwargs):
        if len(self.root.handlers) == 0:
            self.basicConfig()
        self.root._log(self.SUCCESS, msg, args, **kwargs)

    def __getattr__(self, item):
        if item != "__getattr__":
            if item in self.__dict__:
                return self.__dict__[item]
        if item == "getLogger":
            return self.getLogger
        elif item == "success":
            return self.success
        elif item == "SUCCESS":
            return self.SUCCESS
            # end if
            pass
        else:
            return getattr(_logging, item)
            # end def
# end class


logging = _LoggingWrapper()


class LevelByNameFilter(object):
    def __init__(self, root=logging.WARNING, debug=None, info=None, success=None, warning=None, error=None,
                 critical=None, by_level=None):
        """
        A filter where you specify logging levels bound to names (package names, as known from importing)

        :param root: level the root should have to be logged. None to disable.

        :param debug: all loggers which should log debug and above.
        :param info: all loggers which should log info and above.
        :param success: all loggers which should log success and above.
        :param warning: all loggers which should log warning and above.
        :param error: all loggers which should log error and above.
        :param critical: all loggers which should log critical and above.

        :param by_level: a dict with levels as a key, and names to log as value.
                        Example: {10: "__main__", 20: "a.b.c", 30: ["a.b.d", "a.b.e"], logging.WARNING: "a"}
        """
        self.mapping = dict()
        if root:
            if isinstance(root, str):
                root = logging.getLoglevelInt(root)
            assert isinstance(root, int)
            self.mapping[""] = root
        # end
        level = logging.DEBUG
        self.parse_argument(debug, logging.DEBUG)
        self.parse_argument(info, logging.INFO)
        self.parse_argument(success, logging.SUCCESS)
        self.parse_argument(warning, logging.WARNING)
        self.parse_argument(error, logging.ERROR)
        self.parse_argument(critical, logging.CRITICAL)

        if by_level:
            assert isinstance(by_level, dict)
            for level, files in by_level.items():
                self.parse_argument(files, level)
                # end for
                # end if

    # end def

    def parse_argument(self, argument, level):
        if argument:
            if isinstance(argument, tuple):
                argument = list(argument)
            if not isinstance(argument, list):
                argument = [argument]
            # end if
            assert isinstance(argument, list)
            for part in argument:
                if isinstance(part, (list, tuple)):
                    argument.extend(part)
                elif not isinstance(part, str):
                    raise TypeError("argument {val!r} is type {type}, should be str.".format(val=part, type=type(part)))
                elif "," in part:
                    argument.append(part.split(","))
                else:
                    self.mapping[part.strip() + "."] = level
                    # end if
                    # end for
                    # end if

    # end def

    def filter(self, record):
        if not self.mapping:
            return False  # allow
        # end if

        name = record.name + "."
        mapping_path = ""  # default is "" = root

        for k in self.mapping:
            if name.startswith(k):
                if len(mapping_path) < len(k):  # we got a longer path. longer = more specific.
                    mapping_path = k
                    # end if
                    # end if
        # end for

        if mapping_path in self.mapping:  # e.g. root "" is not specified.
            level = self.mapping[mapping_path]
            return record.levelno >= level
        # end if
        return False
        # end def
# end class

        # # Test code to get a threaded logger:
        # from luckydonaldUtils.logger import logging;import threading; from time import sleep;
        # def lel():
        #     logger.debug(threading.current_thread().name)
        #     logging.test_logger_levels(),logger.critical("littlepip is\nBEST\npony!")
        # # end def
# logger = logging.add_colored_handler(level=logging.DEBUG, date_formatter="%Y-%m-%d %H:%M:%S");logging.add_colored_handler(level=logging.DEBUG); lel();sleep(1);thread=threading.Thread(target=lel);thread.start();thread.join()



# -*- coding: utf-8 -*-

import sys
from ..dependencies import import_or_install
from ..logger import logging
import datetime as datetime

try:
    from pytz import timezone
except ImportError:  # pragma nocover
    timezone = import_or_install("pytz.timezone", "pytz")
# end try

__author__ = 'luckydonald'
__all__ = ["get_local_time", "loop_seconds_between_times"]

logger = logging.getLogger(__name__)

if sys.version < "3":
    def get_local_time(time, time_zone):
        return _get_local_time(time, time_zone)


    def loop_seconds_between_times(datetime1, datetime2):
        return _loop_seconds_between_times(datetime1, datetime2)
else:
    from .py3 import get_local_time, loop_seconds_between_times
# end if


def _get_local_time(time, time_zone):
    if time_zone is None:
        time_zone = "Europe/Berlin"  # todo: config
    return timezone('UTC').localize(time).astimezone(timezone(time_zone))


def _loop_seconds_between_times(datetime1, datetime2):
    delta = datetime2 - datetime1
    seconds = delta.total_seconds()
    for x in range(0, int(seconds)):
        yield (datetime1 + datetime.timedelta(seconds=x))
    yield datetime2  # so loop_seconds_between_times(dat,dat) is at least called once with dat.



# -*- coding: utf-8 -*-
from ..functions import deprecated
from ..logger import logging

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)

__all__ = [
    "basics", "mime", "name", "temp", "tree",
    # for downwards compatibility:
    'mkdir_p', 'open_file_folder', 'open_folder', 'get_byte_mime', 'get_file_mime', 'guess_extension',
    'get_file_suffix', 'do_a_filename', 'gettempdir'
]
__IMPORT_CHANGED_MESSAGE = (
    "Import path changed. Please import {func_name} from luckydonaldUtils.files.{where} instead. "
    "This import will be removed in the future."
)


@deprecated(__IMPORT_CHANGED_MESSAGE.format(func_name="mkdir_p", where="basics"))
def mkdir_p(path):
    from .basics import mkdir_p
    return mkdir_p(path)


# end def


@deprecated(__IMPORT_CHANGED_MESSAGE.format(func_name="open_file_folder", where="basics"))
def open_file_folder(file_path):
    from .basics import open_file_folder
    return open_file_folder(file_path)


# end def


@deprecated(__IMPORT_CHANGED_MESSAGE.format(func_name="open_folder", where="basics"))
def open_folder(folder_path):
    from .basics import open_folder
    return open_folder(folder_path)


# end def


@deprecated(__IMPORT_CHANGED_MESSAGE.format(func_name="get_byte_mime", where="mime"))
def get_byte_mime(bytes):
    from .mime import get_byte_mime
    return get_byte_mime(bytes)


# end def


@deprecated(__IMPORT_CHANGED_MESSAGE.format(func_name="get_file_mime", where="mime"))
def get_file_mime(file_path=None, file_url=None):
    from .mime import get_file_mime
    return get_file_mime(file_path=file_path, file_url=file_url)


# end def


@deprecated(__IMPORT_CHANGED_MESSAGE.format(func_name="guess_extension", where="mime"))
def guess_extension(mime):
    from .mime import guess_extension
    return guess_extension(mime)


# end def


@deprecated(__IMPORT_CHANGED_MESSAGE.format(func_name="get_file_suffix", where="mime"))
def get_file_suffix(file_path=None, file_url=None):
    from .mime import get_file_suffix
    return get_file_suffix(file_path=file_path, file_url=file_url)


# end def


@deprecated(__IMPORT_CHANGED_MESSAGE.format(func_name="do_a_filename", where="name"))
def do_a_filename(input_file_name):
    from .name import do_a_filename
    return do_a_filename(input_file_name)


# end def


@deprecated(__IMPORT_CHANGED_MESSAGE.format(func_name="gettempdir", where="temp"))
def gettempdir(temp_folder_name="luckydonald-utils"):
    from .temp import gettempdir
    return gettempdir(temp_folder_name=temp_folder_name)

# end def



# -*- coding: utf-8 -*-

import pip
from luckydonaldUtils.logger import logging

__author__ = 'luckydonald'
logger = logging.getLogger(__name__)


def pip_install(*args):
    return pip.main(["install"].append(args))



"""
    Only one Instance of the class.
    Like in Java .getInstance()


    Make a class Singleton:

    Python 3:
    apply with keyword argument in the base-class list:
    >>> class Example(object, metaclass = Singleton):
    >>>    pass

    Python 2.7:
    apply like
    >>> class Exampe(object):
    >>>    __metaclass__ = Singleton
    >>>    pass


    Usage of the class:

    At the first call you have to supply the parameters for class, as usual.
    Afterwards you can omit them, they will be ignored if given.

    Create the instance:
    >>> whatever = Example("arguments", "go", "here", as_well_as="keyword args")

    Use a already created instance:
    >>> something = Example()  # note: No args or kwargs required here.
"""


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            args = args if args else []
            kwargs = kwargs if kwargs else {}
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]
        # end def



import errno
import os
import tempfile


def gettempdir(temp_folder_name="luckydonald-utils"):
    """gets or creates a folder in the temporary files of the system."""
    temp_dir = tempfile.gettempdir()
    temp_dir = os.path.join(temp_dir, temp_folder_name)
    # py3
    # os.makedirs(temp_dir, exist_ok=True)  # don't raise errors if existent.
    # py2/3 exist_ok workaround
    try:
        os.makedirs(temp_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:  # don't raise errors if existent.
            raise
    # end exist_ok workaround
    return temp_dir



# -*- coding: utf-8 -*-
from random import randint
from .quotes import get_quote

__author__ = 'luckydonald'


def get_headers(language="en"):
    headers = {}
    headers["Server"] = ["iPod Touch, iOS 2.3.2", "Banana 0.2", "KONICHIWA/1.0", "'; DROP TABLE servertypes; -"][randint(0,3)]
    headers["X-Powered-By"] = ["Magical Ponies", "Rats in our Basement", "Unicorns", "Friendship", "TONS OF SUGAR", "coffee", "Bananas and Rum"][randint(0,6)]
    headers["X-Best-Pony"] = "Littlepip"
    headers["X-Answer"] = "42"
    headers["X-Never-Gonna"] = "Give you up."
    headers["X-Nananana"] = "Batman!"
    headers["X-Message"] = get_quote(language)
    return headers



# -*- coding: utf-8 -*-

import logging
import warnings

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)


def assert_type_or_raise(value, expected_type_clazz_or_tuple, *more_clazzes, exception_clazz=TypeError,
                         parameter_name=None):
    """
    A better `assert(isinstance(a, B)` because it supports `None` (as well as some other types except tuple or list), and an nice exception text.

    :param value: The variable to check
    :param expected_type_clazz_or_tuple: a class or value  or  list/tuple of that
    :param *more_clazzes: even more classes or values
    :param exception_clazz: type of the exception raised. Defaults to :class:`TypeError`
    :param parameter_name: setting will add information that given parameter was wrong, instead of just saying "Should be type"  

    :raises: exception_clazz
    :return: Nothing


    Examples/Tests:

    Check if None:

        >>> val = None
        >>> assert_type_or_raise(val, None)

        >>> assert_type_or_raise(val, [None])

    Check multible:

        >>> assert_type_or_raise(val, [None, bool, int])

        >>> assert_type_or_raise(val, (None, bool, int))

        >>> assert_type_or_raise(val, None, bool, int)

        >>> assert_type_or_raise(val, [None, bool], int)

        >>> assert_type_or_raise(val, None, [bool, int])

        >>> assert_type_or_raise(1, [None, bool, int])

        >>> assert_type_or_raise(True, [None, bool, int])

        >>> assert_type_or_raise(1, [None, bool, 1])

        >>> assert_type_or_raise(True, [None, True, int])
        
        >>> assert_type_or_raise("lel", [None, True, int, "lel"])
        
        >>> assert_type_or_raise(True, [None, True, int, "lel"])
        
        >>> assert_type_or_raise(True, [None, True, int, "lel"], exception_clazz=ValueError)
        
        >>> assert_type_or_raise(True, [None, True, int, "lel"], parameter_name="foobar")
        

    Failing asserts:

        >>> assert_type_or_raise("True", [None, True, int])
        Traceback (most recent call last):
        ...
        TypeError: Should be one of the types [None, True, <class 'int'>], but is type <class 'str'>: 'True'

        >>> assert_type_or_raise(None, [True, int])
        Traceback (most recent call last):
        ...
        TypeError: Should be one of the types [True, <class 'int'>], but is type <class 'NoneType'>: None


        >>> assert_type_or_raise("lelz", [None, True, int, "lel"])
        Traceback (most recent call last):
        ...
        TypeError: Should be one of the types [None, True, <class 'int'>, 'lel'], but is type <class 'str'>: 'lelz'


    Custom Exception Type:

        >>> assert_type_or_raise(None, [True, int], exception_clazz=ValueError)
        Traceback (most recent call last):
        ...
        ValueError: Should be one of the types [True, <class 'int'>], but is type <class 'NoneType'>: None


    Parameter name supplied:
        
        >>> assert_type_or_raise(None, [True, int], parameter_name="foobar")
        Traceback (most recent call last):
        ...
        TypeError: The parameter foobar should be one of the types [True, <class 'int'>], but is type <class 'NoneType'>: None
    
    """
    if expected_type_clazz_or_tuple is None and value is None:
        return  # None == None
    # end if
    if isinstance(expected_type_clazz_or_tuple, tuple):
        expected_type_clazz_or_tuple = list(expected_type_clazz_or_tuple)
    if not isinstance(expected_type_clazz_or_tuple, list):
        expected_type_clazz_or_tuple = [expected_type_clazz_or_tuple]
    # end if
    expected_type_clazz_or_tuple.extend(more_clazzes)
    parameter_string = "the parameter {name} ".format(name=parameter_name) if parameter_name else ""
    for list_item in expected_type_clazz_or_tuple:
        if isinstance(list_item, (list, tuple)):
            expected_type_clazz_or_tuple.extend(list(list_item))
            continue
        if list_item is None:
            if value is None:
                assert value is None
                break  # is valid
            else:
                continue  # is None, but value is not.
                # end if
        if isinstance(list_item, type):
            if isinstance(value, list_item):
                assert isinstance(value, list_item)
                break  # is valid
                # end if
        else:
            if value is list_item or value == list_item:
                assert value is list_item or value == list_item
                break  # is valid
                # end if
                # end if
    else:  # for: -> is not valid, no break was used
        if len(expected_type_clazz_or_tuple) != 1:
            exception_text = (
                "{parameter_string}should be one of the types [{expected_type}], but is type {real_type}: {real_value!r}".format(
                    parameter_string=parameter_string, real_type=type(value), real_value=value,
                    expected_type=", ".join(repr(x) for x in expected_type_clazz_or_tuple)
                )
            )
        else:
            exception_text = (
                "{parameter_string}should be of type {expected_type!r}, but is type {real_type}: {real_value!r}".format(
                    parameter_string=parameter_string, expected_type=expected_type_clazz_or_tuple,
                    real_type=type(value), real_value=value
                )
            )
        # end if
        exception_text = exception_text[0].upper() + exception_text[1:]
        raise exception_clazz(exception_text)
        # end for
# end def assert_or_raise


def assert_or_raise(*args, **kwargs):
    warnings.warn("The `luckydonaldUtils.exceptions.assert_or_raise` method is deprecated, "
                  "use `assert_type_or_raise` instead", DeprecationWarning, 2)
    logger.warn("Deprecated: `luckydonaldUtils.exceptions.assert_or_raise` was renamed to `assert_type_or_raise`.")
    return assert_type_or_raise(*args, **kwargs)

# end def



# -*- coding: utf-8 -*-
from luckydonaldUtils.logger import logging

try:
    from werkzeug.exceptions import HTTPException
    from werkzeug.routing import MapAdapter
    from flask.globals import _app_ctx_stack, _request_ctx_stack
    from flask import url_for, request

except ImportError:  # pragma nocover
    from ..dependencies import import_or_install

    MapAdapter = import_or_install("werkzeug.routing.MapAdapter", "werkzeug")
    HTTPException = import_or_install("werkzeug.exceptions.HTTPException", "werkzeug")
    _app_ctx_stack = import_or_install("flask.globals._app_ctx_stack", "flask")
    url_for, request = import_or_install("flask.url_for", "flask"), import_or_install("flask.request", "flask")
    _request_ctx_stack = import_or_install("flask.globals._request_ctx_stack", "flask")
# end try


__author__ = 'luckydonald'
logger = logging.getLogger(__name__)


def route_for(url, is_full_url=False):
    """
    Basically `flask.url_for` in reverse.
    You give a url, it gives you the route and the required arguments.

    Example:
    >>> @app.route('/foobar/<argument>')
    ... def some_test(argument):
    ...     pass
    ... # end def

    >>> route_for("https://example.com/foobar/lel", is_full_url=True)



    :param url: The url you want the route for.
    :param is_full_url: If true it will try to remove `{adapter.url_scheme}://{adapter.server_name}` from the beginning,
                        and return `None` if it fails.
    :return: you get a tuple in the form ``(endpoint, arguments)`` if there is a match, else if anything goes wrong, ``None``.
    :rtype: None | tuple of (str, list of str)
    """
    adapter = _app_ctx_stack.top.url_adapter if _app_ctx_stack and _app_ctx_stack.top and _app_ctx_stack.top.url_adapter else _request_ctx_stack.top.url_adapter
    assert isinstance(adapter, MapAdapter)
    if is_full_url:
        if not url.startswith(adapter.url_scheme + '://'):
            logger.debug("Url doesn't start with '{}://'.".format(adapter.url_scheme))
            return None
        url = url[len(adapter.url_scheme) + 3:]
        # end if
        if not url.startswith(adapter.server_name):
            logger.debug("Url doesn't start with '{}://{}'.".format(adapter.url_scheme, adapter.server_name))
            return None
        # end if
        url = url[len(adapter.server_name):]
        if "?" in url:
            url = url.split("?", 1)[0]
            logger.debug("Removed query segment.")
        # end if
        if "#" in url:
            url = url.split("#", 1)[0]
            logger.debug("Removed hash segment.")
        # end if
    # end if

    try:
        return adapter.match(url, method=request.method, return_rule=False)
    except HTTPException as e:
        logger.debug("Route not found: {}".format(e))
        return None
    # end try


# end def


def get_safe_next(url=None, full_url_only=True, allow_self=False, old_referrer_must_match=False,
                  generate_url=True):
    """
    Checks if the specified ``url`` in like a ``?next=`` parameter is part of our registered routes,
    and thus seems safe to redirect to.

    Additionally you can choose:

    - if you want to require the value to be a full url, with protocol, domain, etc. (``full_url_only``),
    - if you want to allow redirect to the same page again which might cause a redirect loop (``allow_self``),
    - if you need the http referrer the browser sent to match the next ``url``,
      i.e. it goes back to the previous url (``old_referrer_must_match``)

    :param url: The value for the next url. Defaults to ``request.args.get('next')``.
    :param full_url_only: If the next url should be a full url, with ``http://www.domain.com/`` and whatnot.
    :param allow_self: If redirect to self is valid. Default: ``False``.

    :param old_referrer_must_match: Default: ``False``. Don't check referrer.
                                    ``True`` to require it to resolve to the same endpoint as ``next`` does.
                                    A ``str`` to require it to match that endpoint.

    :type  old_referrer_must_match: bool|str

    :param generate_url: If we should return an url string (``True``), or just the (``route``, ``params``) tuple (``False``).
    :type  generate_url: bool

    :return: If anything goes wrong, or you get no match, ``None`` is returned. Else, if ``generate_url`` is ``True``, you get an url string you can redirect to, or if ``generate_url`` is ``False`` you get a tuple in the form ``(endpoint, arguments)``.
    :rtype: None | str | tuple of (str, dict)
    """
    if not url:
        url = request.args.get('next', None)
    # end if
    if not url:
        logger.debug('No next parameter.')
        return None
    # end if
    param_route = route_for(url, is_full_url=full_url_only)
    if not param_route:
        logger.debug('No route found for next parameter.')
        return None
    # end if
    if old_referrer_must_match:
        # needed_referrer_endpoint  is either the string endpoint in old_referrer_must_match,
        # or the route calculated for the `next` param, param_route[0]
        if not request.referrer:
            logger.debug('No referrer data.')
            return None
        # end if

        needed_referrer_endpoint = old_referrer_must_match if isinstance(old_referrer_must_match, str) else param_route[
            0]
        referrer_route = route_for(request.referrer, is_full_url=True)

        if referrer_route[0] != needed_referrer_endpoint:
            logger.debug('Referrer endpoint {!r} does not fit needed endpoint {!r}.'.format(
                referrer_route[0], needed_referrer_endpoint
            ))
            return None
        # end if
    # end if
    if not allow_self and param_route[0] == request.endpoint:
        # don't redirect to self
        logger.debug('Redirect to self is not allowed.')
        return None
    # end if
    logger.debug('Next url ok: {!r}'.format(param_route[0]))
    if generate_url:
        return url_for(param_route[0], **param_route[1])
    # end if
    return param_route
# end def



# -*- coding: utf-8 -*-
"""
Tools to use with django's HttpRequests.
"""
try:
    from django.middleware.csrf import CsrfViewMiddleware
except ImportError:  # pragma nocover
    from ..dependencies import import_or_install

    CsrfViewMiddleware = import_or_install("django.middleware.csrf.CsrfViewMiddleware", "django")
# end try


__author__ = 'luckydonald'


def check_csrf(request):
    """
    This checks if the csrf check was passed.

    :param request: the request
    :type  request: django.http.HttpRequest

    :return: If it did pass the check.
    :rtype : bool
    """
    reason = CsrfViewMiddleware().process_view(request, None, (), {})
    if reason:
        # CSRF failed
        return False
    return True


# end def

def GET_to_bool(request, key):
    """
    Parses a GET parameter in the request as bool.
    "true" becomes True, "false" becomes False, "null" becomes None.

    :param request: the request
    :type  request: django.http.HttpRequest

    :param key: the key in request.GET
    :type  key: str

    :return: Boolean or None
    :rtype : bool | None
    """
    clone_manual = request.GET[key]
    assert clone_manual in ["true", "false", "null"]
    clone_manual_bool = None
    if clone_manual == "true":
        clone_manual_bool = True
    if clone_manual == "false":
        clone_manual_bool = False
    return clone_manual_bool



# -*- coding: utf-8 -*-
try:
    from .encoding import binary_type, text_type
except (ImportError, SystemError):
    # non-relative imports to enable doctests
    from luckydonaldUtils.encoding import binary_type, text_type
# end if

__author__ = 'luckydonald'

CHARS_UNESCAPED = ["\\", "\n", "\r", "\t", "\b", "\a", "'"]
CHARS_ESCAPED = ["\\\\", "\\n", "\\r", "\\t", "\\b", "\\a", "\\'"]


def text_split(text, limit, max_parts=None):
    """
    Split a text, each part less or equal to `limit`. Prefer splitting at linebreaks, sentences and lastly words.

    :type text: str
    :type limit: int
    :param max_parts: If not None: In how many parts it should be split.
                      The last element might be longer as the given limit.
    :type  max_parts: None | int
    :return:
    """
    if not isinstance(text, str):
        raise TypeError("text is not str but {}".format(type(text)))
    assert isinstance(limit, int)
    assert max_parts is None or isinstance(max_parts, int)
    text = text.strip()
    if len(text) == 0 or limit == 0:
        return []
    if max_parts == 1:
        return [text]
    assert max_parts is None or max_parts > 1
    # LinkRanges lnkRanges = textParseLinks(text, TextParseLinks | TextParseMentions | TextParseHashtags);
    # currentLink = 0
    # lnkCount = lnkRanges.size()
    half = limit // 2
    good_level = 0
    good = 0
    end = len(text)
    unicode_i = 0
    for i in range(0, end):
        char = text[i]
        # while (currentLink < lnkCount && ch >= lnkRanges[currentLink].from + lnkRanges[currentLink].len) {
        # 	++currentLink;
        #  }
        # bool inLink = (currentLink < lnkCount) && (ch > lnkRanges[currentLink].from) &&
        #   (ch < lnkRanges[currentLink].from + lnkRanges[currentLink].len);
        if unicode_i > half:
            if char == "\n":
                if i + 1 < end and text[i + 1] == "\n" and good_level <= 7:
                    good_level = 7
                    good = i
                elif good_level <= 6:
                    good_level = 6
                    good = i
            elif char.isspace():
                if is_sentence_end(char) and good_level <= 5:
                    good_level = 5
                    good = i
                elif is_sentence_part_end(text[i - 1]) and good_level <= 4:
                    good_level = 4
                    good = i
                elif good_level <= 3:
                    good_level = 3
                    good = i
            elif is_word_separator(text[i - 1]) and good_level <= 2:
                good_level = 2
                good = i
            elif good_level <= 1:
                good_level = 1
                good = i
                # end if
        # end if
        if unicode_i >= limit:
            if max_parts == 1:
                return [text.strip()]
            first_text = text[:good]
            text = text[good:]
            parts = [first_text.strip(), ]
            if max_parts is None:
                parts.extend(text_split(text, limit))
            elif max_parts > 1:
                parts.extend(text_split(text, limit, max_parts=max_parts - 1))
            else:
                raise ArithmeticError("max_split reached illegal state: {}".format(max_parts))
            return parts
        # end if
        if char in CHARS_UNESCAPED:
            unicode_i += 2
        else:
            unicode_i += len(char.encode('utf-8'))
    # end for
    return [text]
# end def


def is_sentence_end(char):
    return char in ".?!"


# end def


def is_sentence_part_end(char):
    return char in ",:;"


# end def


def is_word_separator(char):
    return char in [
        " ", "\n", ".", ",", "?", "!", "@", "#", "$", ":", ";", "-", "<", ">", "[", "]",
        "(", ")", "{", "}", "=", "/", "+", "%", "&", "^", "*", "'", "\"", "`", "´", "„", "”", "~", "|"
    ]


# end def


def split_in_parts(string, parts, strict=False):
    """
    Splits a string in `parts` amount of pieces.
    If `string` is a :class:`list`, which has not exactly `parts` items, it will be joined and split afterwards.

    Examples:

      >>> split_in_parts('abc', 3)
      ['a', 'b', 'c']
      >>> split_in_parts('abcd', 3)
      ['a', 'bc', 'd']
      >>> split_in_parts('abcde', 3)
      ['ab', 'c', 'de']
      >>> split_in_parts('abcdef', 3)
      ['ab', 'cd', 'ef']
      >>> split_in_parts("abcdefgh", 3)
      ['abc', 'de', 'fgh']
      >>> split_in_parts('║╠╚ ', 4)
      ['║', '╠', '╚', ' ']
      >>> split_in_parts('ab', 4)
      ['', 'a', 'b', '']
      >>> split_in_parts('ab', 3)
      ['a', '', 'b']
      >>> split_in_parts('║╟╙╴⚠', 5)
      ['║', '╟', '╙', '╴', '⚠']
      
      >>> split_in_parts(['a','b','c'], 3)  # list with right amount of parts, unchanged
      ['a', 'b', 'c']
      >>> split_in_parts(['aa','b','c'], 3)  # list with  right amount of parts, unchanged
      ['aa', 'b', 'c']
      >>> split_in_parts(['a','b','c','d'], 3)  # list with wrong amount of parts, joined + splitted
      ['a', 'bc', 'd']
      >>> split_in_parts(['aa','b','c','d'], 3)  # list with wrong amount of parts, joined + splitted
      ['aa', 'b', 'cd']
      
      >>> split_in_parts([], 0)  # empty list with right amount of parts, unchanged
      []
      >>> split_in_parts([], 1)  # empty list with wrong amount of parts, empty string generated
      ['']
      >>> split_in_parts([], 2)  # empty list with wrong amount of parts, empty string generated
      ['', '']
      
      >>> split_in_parts("abc", 3, strict=True)  # strict mode, fitting `parts` parameter
      ['a', 'b', 'c']
      >>> split_in_parts("abc", 4, strict=True)  # strict mode, wrong `parts` parameter
      Traceback (most recent call last):
      ...
      ValueError: In strict mode you need a string which can be split in 4 equal pieces.
      >>> split_in_parts("abcdef", 2, strict=True)  # strict mode, wrong `parts` parameter
      ['abc', 'def']
      
      >>> split_in_parts(['a','b','c'], 3, strict=True)  # list with right amount of parts, unchanged
      ['a', 'b', 'c']
      >>> split_in_parts(['aa','b','c'], 3, strict=True)  # list with  right amount of parts, but not equal length
      Traceback (most recent call last):
      ...
      ValueError: In strict mode you need a list where all 3 items have equal length.
      >>> split_in_parts(['aaa','bbb','ccc','ddd'], 3, strict=True)  # list with wrong amount of parts, joined + splitted
      ['aaab', 'bbcc', 'cddd']
      >>> split_in_parts(['aa','b','c','d'], 3, strict=True)  # list with wrong amount of parts, joined + splitted
      Traceback (most recent call last):
      ...
      ValueError: In strict mode you need a string which can be split in 3 equal pieces.
    
      >>> split_in_parts([], 0, strict=True)  # empty list with right amount of parts, unchanged
      []
      >>> split_in_parts([], 1, strict=True)  # empty list with wrong amount of parts, empty string generated
      ['']
      >>> split_in_parts([], 2, strict=True)  # empty list with wrong amount of parts, empty string generated
      ['', '']
      

    :param string: the string to split
    :param parts: how many parts
    :param strict: if it should only have parts which are all same length.

    :return: list of strings
    """
    if isinstance(string, list):
        if len(string) == parts:
            if strict and len(string) != 0:
                length = len(string[0])
                if any(len(item) != length for item in string):  # length wrong in some element
                    raise ValueError("In strict mode you need a list where all {parts} items have equal length.".format(
                        parts=parts
                    ))
                    # end if  length wrong
            # end if strict
            return string
        # end if
        string = "".join(string)
    # end if
    length = len(string)
    slice_size = length / parts
    if strict and length % parts != 0:
        raise ValueError("In strict mode you need a string which can be split in {parts} equal pieces.".format(
            parts=parts
        ))
    # end if
    part_pos = [0]
    for i in range(0, parts):  # parts = 3 -> for [1, 2]
        part_pos.append(round(slice_size * (i + 1)))
    # end for
    # print(part_pos)
    part_pos[1] = round(part_pos[1])
    strings = []
    for i in range(parts):
        strings.append(string[part_pos[i]:part_pos[i + 1]])
    # end for
    return strings
# end def


def escape(string):
    for i in range(0, 7):
        string = string.replace(CHARS_UNESCAPED[i], CHARS_ESCAPED[i])
    return string


# end def


def cut_paragraphs(text, length=200) -> str:
    """
    This limits a paragraph to a given length.
    :param text:
    :param length:
    :return:
    """
    if len(text) == length:
        return text
    short = text[:length]
    last_sentence_ending = max(short.rfind(x) for x in ["\n", "...", ";", ":", "!", "?", "."])
    if last_sentence_ending > 0 and length - last_sentence_ending < 20:  # limit it to max. removing 20 characters,
        # else try something different.
        return short[:last_sentence_ending + 1]
    from .holder import Holder
    h = Holder()
    if h(short.rfind(",", 0, -3)) > 0 and h() < length - 3:
        return short[:h() + 1] + "..."
    if h(short.rfind(" ", 0, -3)) > 0:
        return short[:h()] + "..."
    # fallback if nothing splittable was found.
    return text[:-3] + "..."  # will just cut it.
# end def


def lcut(input_string, part_to_cut):
    """
    Cuts away a `part_to_cut` from the beginning of the `input_string`.
    If your `input_string` is no text type, it will be returned unchanged.
    
    
    Examples: 
        
        >>> lcut("foobar", "foo")
        'bar'
        
        >>> lcut("foobar", "batz")
        'foobar'
        
        >>> lcut(False, "False")
        False
        
        >>> lcut(u"foobar", u"foo") == u"bar"  # unicode
        True
        
        >>> lcut(b"foobar", b"foo") == b"bar"  # binary
        True
        
        >>> lcut(b"foobar", u"foo") == b"bar"  # binary vs unicode  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ...
        TypeError: startswith first arg must be bytes or a tuple of bytes, not str

    
    :param input_string: Your complete string you want to cut stuff from
    :param part_to_cut: What you want to cut away from the start, if found.
    :return: 
    """
    if not isinstance(input_string, (text_type, binary_type)):  # non-text
        return input_string
    # end if
    if input_string.startswith(part_to_cut):
        return input_string[len(part_to_cut):]
    # end if
    return input_string


# end def


def rcut(input_string, part_to_cut):
    """
    Cuts away a `part_to_cut` from the end of the `input_string`.
    If your `input_string` is no text type, it will be returned unchanged.
    
    
    Examples: 
        
        >>> rcut("foobar", "bar")
        'foo'
        
        >>> rcut("foobar", "batz")
        'foobar'
        
        >>> rcut(False, "False")
        False
        
        >>> rcut(u"foobar", u"bar") == u"foo"  # unicode
        True
        
        >>> rcut(b"foobar", b"bar") == b"foo"  # binary
        True
        
        >>> rcut(b"foobar", u"bar") == b"foo"  # binary vs unicode  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ...
        TypeError: endswith first arg must be bytes or a tuple of bytes, not str


    :param input_string: Your complete string you want to cut stuff from
    :param part_to_cut: What you want to cut away from the start, if found.
    :return: 
    """
    if not isinstance(input_string, (text_type, binary_type)):  # non-text
        return input_string
    # end if
    if input_string.endswith(part_to_cut):
        return input_string[:len(part_to_cut)]
    # end if
    return input_string
# end def
